'
'    Copyright Ad Lib Inc., 1990
'
'    This file is part of the Ad Lib Programmer's Manual product and is
'    subject to copyright laws.  As such, you may make copies of this file
'    only for the purpose of having backup copies.  This file may not be
'    redistributed in any form whatsoever.
'
'    If you find yourself in possession of this file without having received
'    it directly from Ad Lib Inc., then you are in violation of copyright
'    laws, which is a form of theft.
'

' File: ADLIB.BAS
' QuickBasic low level sound driver V1.00
'    Adapted from adlib.c, v1.04

' Copyright AdLib Inc., 1990

' March 1990, Dale Glowinski

' This module uses the external routine SndOutput() which is in the
' library ADLIB.QLB.  The assembler source code for this file can be
' found in the file OUTCHIP.ASM.

' In percussive mode, the last 4 voices (SD TOM HH CYMB) are created
' using melodic voices 7 & 8. A noise generator uses channels 7 & 8
' frequency information for creating rhythm instruments. Best result
' are obtained by setting TOM two octaves below middle C and SD 7
' half-tones above TOM.
' In this implementation, of the 4 percussive voices only the TOM pitch
' may vary, with the SD always set 7 half-tones above.  The HH and CYMB
' pitches are fixed.

'Declarations for external routines.
DECLARE SUB SndOutput CDECL (BYVAL regAddr%, BYVAL dataByte%)

'Declarations for this module's routines
DECLARE SUB InitSlotParams ()
DECLARE SUB NoteOff (voice AS INTEGER)
DECLARE SUB NoteOn (voice AS INTEGER, note AS INTEGER)
DECLARE SUB OutFreq (voice AS INTEGER, pitch AS INTEGER)
DECLARE SUB SetGParam (amD AS INTEGER, vibD AS INTEGER, nSel AS INTEGER)
DECLARE SUB SetMode (mode AS INTEGER)
DECLARE SUB SetPitchRange (pR AS INTEGER)
DECLARE SUB SetSlotParam (slot AS INTEGER, param() AS INTEGER, waveSel AS INTEGER)
DECLARE SUB SetUpData ()
DECLARE SUB SetUpFNums ()
DECLARE SUB SetWaveSel (state AS INTEGER)
DECLARE SUB SndSAmVibRhythm ()
DECLARE SUB SndSAttDecay (slot AS INTEGER)
DECLARE SUB SndSAVEK (slot AS INTEGER)
DECLARE SUB SndSetAllPrm (slot AS INTEGER)
DECLARE SUB SndSetPrm (slot AS INTEGER, prm AS INTEGER)
DECLARE SUB SndSFeedFm (slot AS INTEGER)
DECLARE SUB SndSKslLevel (slot AS INTEGER)
DECLARE SUB SndSNoteSel ()
DECLARE SUB SndSSusRelease (slot AS INTEGER)
DECLARE SUB SndWaveSelect (slot AS INTEGER)
DECLARE SUB SoundWarmInit ()
DECLARE FUNCTION SoundColdInit% (port AS INTEGER)
DECLARE FUNCTION BoardInstalled% ()


'-----------------------------------------------------------------
' Number of instrument (patch) parameters for each voice:
CONST insParams = 13

CONST prmKsl = 0
CONST prmMulti = 1
CONST prmFeedBack = 2                  ' use for opr. 0 only
CONST prmAttack = 3
CONST prmSustain = 4
CONST prmStaining = 5                  ' Sustaining ...
CONST prmDecay = 6
CONST prmRelease = 7
CONST prmLevel = 8
CONST prmAm = 9
CONST prmVib = 10
CONST prmKsr = 11
CONST prmFm = 12                       ' use for opr. 0 only
CONST prmWaveSel = 13                  ' wave select
CONST prmAmDepth = 14
CONST prmVibDepth = 15
CONST prmNoteSel = 16
CONST prmPercussion = 17

' percussive voice numbers:
CONST BD = 6
CONST SD = 7
CONST TOM = 8
CONST CYMB = 9
CONST HIHAT = 10

CONST MAXVOLUME = &H7F
CONST MAXBEND = &H3FFF
CONST NOBEND = &H2000

CONST MIDC = 60                     ' MIDI standard mid C
CONST NRNOTES = 96                  ' # of notes we can play on chip
       
CONST TOMPITCH = 36                 'best frequency for tom-tom
CONST TOMTOSD = 7                   '7 half-tones between voice 7 & 8
CONST SDPITCH = (TOMPITCH + TOMTOSD)


'-----------------------------------------------------------------

'Port address of Ad Lib card
DIM SHARED genAddr AS INTEGER

'Pitch bend variation, 1 - 12 half-tones
DIM SHARED pitchRange AS INTEGER

'<> 0 if using card in multiple wave form mode (WSE = 1)
DIM SHARED modeWaveSel AS INTEGER

'Bits used to do note on/off for percussive voices.
DIM SHARED percBits AS INTEGER

'Table of key-on/key-off bits for the percussive voices.
DIM SHARED percMasks(0 TO 4) AS INTEGER
DpercMasks:
   DATA  &H10, &H08, &H04, &H02, &H01

DIM SHARED voiceNote(0 TO 8) AS INTEGER       'pitch of last note-on of each voice
DIM SHARED vPitchBend(0 TO 8) AS INTEGER      'current pitch bend of each voice
DIM SHARED bxRegister(0 TO 8) AS INTEGER      'current val. of &HB0 - &HB8 reg
DIM SHARED lVoiceVolume(0 TO 10) AS INTEGER   'volume for each of 11 logical voices

'9 or 11, depending on whether in melodic or percussive mode.
DIM SHARED modeVoices AS INTEGER

'definition of the ELECTRIC-PIANO voice (opr0 & opr1)
DIM SHARED pianoParamsOp0(0 TO insParams) AS INTEGER
DpianoParamsOp0:
   DATA  1, 1, 3, 15, 5, 0, 1, 3, 15, 0, 0, 0, 1, 0
DIM SHARED pianoParamsOp1(0 TO insParams) AS INTEGER
   DATA  0, 1, 1, 15, 7, 0, 2, 4, 0, 0, 0, 1, 0, 0

'definition of default percussive voices:
DIM SHARED bdOpr0(0 TO insParams) AS INTEGER
DpercParams:
   DATA 0,  0, 0, 10, 4, 0, 8, 12, 11, 0, 0, 0, 1, 0
DIM SHARED bdOpr1(0 TO insParams) AS INTEGER
   DATA 0,  0, 0, 13, 4, 0, 6, 15,  0, 0, 0, 0, 1, 0
DIM SHARED sdOpr(0 TO insParams) AS INTEGER
   DATA 0, 12, 0, 15, 11, 0, 8,  5,  0, 0, 0, 0, 0, 0
DIM SHARED tomOpr(0 TO insParams) AS INTEGER
   DATA  0,  4, 0, 15, 11, 0, 7,  5,  0, 0, 0, 0, 0, 0
DIM SHARED cymbOpr(0 TO insParams) AS INTEGER
   DATA 0,  1, 0, 15, 11, 0, 5,  5,  0, 0, 0, 0, 0, 0
DIM SHARED hhOpr(0 TO insParams) AS INTEGER
   DATA 0,  1, 0, 15, 11, 0, 7,  5,  0, 0, 0, 0, 0, 0

'all the parameters of slots...
DIM SHARED paramSlot(0 TO 17, 0 TO insParams) AS INTEGER

DIM SHARED amDepth AS INTEGER              'chip global parameters ..
DIM SHARED vibDepth AS INTEGER             '...
DIM SHARED noteSel AS INTEGER              '...
DIM SHARED percussion AS INTEGER           'percussion mode parameter

'Slot numbers as a function of the voice and the operator for when in
'melodic mode.
DIM SHARED slotMVoice(0 TO 8, 0 TO 1) AS INTEGER
DslotMVoice:
   'voices 0 - 8
   DATA 0, 3
   DATA 1, 4
   DATA 2, 5
   DATA 6, 9
   DATA 7, 10
   DATA 8, 11
   DATA 12, 15
   DATA 13, 16
   DATA 14, 17

'Slot numbers as a function of the voice and the operator for when in
'percussive mode.  (255 indicates that there is only one slot.)
DIM SHARED slotPVoice(0 TO 10, 0 TO 1) AS INTEGER
DslotPVoice:
   'voices 0 - 5, bass drum
   DATA 0, 3
   DATA 1, 4
   DATA 2, 5
   DATA 6, 9
   DATA 7, 10
   DATA 8, 11
   DATA 12, 15
   'snare (slot 16), tom (slot 14), cymbol (slot 17), hihat (slot 13)
   DATA 16, 255
   DATA 14, 255
   DATA 17, 255
   DATA 13, 255

'This table gives the offset of each slot within the chip given the slot.
DIM SHARED offsetSlot(0 TO 17) AS INTEGER
DoffsetSlot:
   DATA 0,  1,  2,  3,  4,  5
   DATA 8,  9, 10, 11, 12, 13
   DATA 16, 17, 18, 19, 20, 21

'This table indicates if the slot is a modulator (0) or a carrier (1).
DIM SHARED carrierSlot(0 TO 17) AS INTEGER
DcarrierSlot:
   DATA 0, 0, 0
   DATA 1, 1, 1
   DATA 0, 0, 0
   DATA 1, 1, 1
   DATA 0, 0, 0
   DATA 1, 1, 1

'This table gives the voice number associated with each slot when in
'melodic mode.
DIM SHARED voiceMSlot(0 TO 17) AS INTEGER
DvoiceMSlot:
   DATA 0, 1, 2
   DATA 0, 1, 2
   DATA 3, 4, 5
   DATA 3, 4, 5
   DATA 6, 7, 8
   DATA 6, 7, 8

'This table gives the voice number associated with each slot when in
'percussive mode.
DIM SHARED voicePSlot(0 TO 17) AS INTEGER
DvoicePSlot:
   DATA 0, 1, 2
   DATA 0, 1, 2
   DATA 3, 4, 5
   DATA 3, 4, 5
   'BD = 6, SD = 7, TOM = 8, CYMB = 9, HIHAT = 10
   DATA 6, 10, 8
   DATA 6, 7, 9

'Table of f-numbers to be placed in the frequency register (&HA0 and &HB0)
'given a note.
DIM SHARED freqNums(0 TO 11) AS INTEGER
DfreqNums:
   '    C    C#   D    D#   E    F
   DATA 343, 363, 385, 408, 432, 458
   '    F#   G    G#   A    A#   B
   DATA 485, 514, 544, 577, 611, 647

END

'Test for board's presence in computer.  The routine set up timer-1 on the
'board and waits for it to run out (overflow).  The status register will
'contain specific values before and after this action.  If the status
'register values are not what they should be, then the board is not present.
'Returns 1 if board found, 0 if not.
FUNCTION BoardInstalled%
   DIM t1 AS INTEGER
   DIM t2 AS INTEGER
   DIM i  AS INTEGER

   CALL SndOutput(4, &H60)     'mask T1 & T2
   CALL SndOutput(4, &H80)     'reset IRQ

   t1 = INP(&H388)             'read status register
   CALL SndOutput(2, &HFF)     'set timer-1 latch
   CALL SndOutput(4, &H21)     'unmask & start T1

   FOR i = 0 TO 200            '100 uSec delay for timer-1 overflow
      tmp% = INP(genAddr)
   NEXT

   t2 = INP(&H388)             'read status register
   CALL SndOutput(4, &H60)
   CALL SndOutput(4, &H80)

   IF ((t1 AND &HE0) = 0 AND (t2 AND &HE0) = &HC0) THEN
      BoardInstalled = 1
   ELSE
      BoardInstalled = 0
   END IF

END FUNCTION

'Set up the default instrument parameters.
'   In melodic mode, initialize all voices to electric-pianos.
'   In percussive mode, initialize the first 6 voices to electric-pianos
'   and the percussive voices to their default timbres.
SUB InitSlotParams
   DIM i AS INTEGER

   FOR i = 0 TO 17
      IF (carrierSlot(i) <> 0) THEN
	 CALL SetSlotParam(i, pianoParamsOp1(), 0)
      ELSE
	 CALL SetSlotParam(i, pianoParamsOp0(), 0)
      END IF
   NEXT

   IF (percussion <> 0) THEN
      CALL SetSlotParam(12, bdOpr0(), 0)
      CALL SetSlotParam(15, bdOpr1(), 0)
      CALL SetSlotParam(16, sdOpr(), 0)
      CALL SetSlotParam(14, tomOpr(), 0)
      CALL SetSlotParam(17, cymbOpr(), 0)
      CALL SetSlotParam(13, hhOpr(), 0)
   END IF
END SUB

'Stop playing the note which was started in 'NoteOn'.
SUB NoteOff (voice AS INTEGER)
   IF (voice < BD OR (percussion = 0 AND voice < 9)) THEN
      'Set the key-on bit to zero to do a note off.  The other information
      'in the register must be preserved as the note does not become silent
      'immediately, but begins decaying at this time.
      bxRegister(voice) = bxRegister(voice) AND (NOT &H20)
      CALL SndOutput(&HB0 + voice, bxRegister(voice))
   ELSEIF (percussion <> 0 AND voice <= HIHAT) THEN
      percBits = percBits AND (NOT percMasks(voice - BD))
      SndSAmVibRhythm
   END IF
END SUB

'Start a note playing.
SUB NoteOn (voice AS INTEGER, pitch AS INTEGER)
  
   'A note-on with a pitch of zero is equivalent to a note-off.
   IF (pitch <= 0) THEN
      voiceNote(voice) = 0
      NoteOff (voice)
      EXIT SUB
   END IF

   IF (voice < BD OR (percussion = 0 AND voice < 9)) THEN
      'This is a melodic voice.
      voiceNote(voice) = pitch
      CALL OutFreq(voice, pitch)
  
   ELSEIF (percussion <> 0 AND voice <= HIHAT) THEN
      'This is a percussive voice.
      'Of the 5 percussive voices, only the BD and the TOM may change in
      'frequency.  The SD remains at a fixed interval above the TOM.
      IF (voice = BD) THEN
	 voiceNote(BD) = pitch
	 CALL OutFreq(BD, pitch)
      ELSEIF (voice = TOM) THEN
	 IF (voiceNote(TOM) <> pitch) THEN
	    voiceNote(TOM) = pitch
	    voiceNote(SD) = pitch + TOMTOSD
	    CALL OutFreq(TOM, pitch)
	    CALL OutFreq(SD, pitch)
	 END IF
      END IF
     
      percBits = percBits OR percMasks(voice - BD)
     
      SndSAmVibRhythm     'Does note-on for percussive voices.
   END IF
END SUB

'Write the appropriate values in registers &HA0 and &HB0 in order to set
'the frequency corresponding to the passed note.  The calling routine is
'responsable for verifying the 'pitch' is at least 12 and that 'voice' is
'is a valid voice number.
'
SUB OutFreq (voice AS INTEGER, pitch AS INTEGER)
   DIM t1 AS INTEGER
   DIM effNbr AS INTEGER

   'Integer division by 12 gives the octave (referred to as BLOCK information
   'in the manual).
   octave = (pitch \ 12) - 1

   'The remainder of dividing by 12 gives the half-tone within the octave.
   'The freqNums table gives the value to place in the register given the
   'half-tone.
   effNbr = freqNums(pitch MOD 12)
  
   IF (vPitchBend(voice) <> &H2000) THEN
      'Do a pitch bend.
      'pitchRange is the maximum interval for a pitch bend.
      'vPitchBend(voice) is the amount of pitch bend.
      DIM n AS LONG
     
      'Below, the maximum interval is calculated in terms of fnums, and
      'then a fraction of this is added to the base value (effNbr).
      IF (vPitchBend(voice) > &H2000) THEN
	 'Make "n" the interval, in f-numbers, between pitch and
	 '(pitch + pitchRange).
	 n = freqNums((pitch + pitchRange) MOD 12)
	 IF (n < effNbr) THEN n = n * 2
	 n = n - effNbr
	
	 'Add a fraction of the interval to effNbr.
	 effNbr = effNbr + ((n * (vPitchBend(voice) - &H2000)) \ &H2000)
	
	 'If effNbr exceeds its maximum possible value, bring it into range
	 'by dividing by two, which lowers the pitch by an octave, and add
	 '1 to the octave to keep the pitch in the correct octave.
	 WHILE (effNbr > 1023)
	    effNbr = effNbr \ 2
	    octave = octave + 1
	 WEND

      ELSE
	 'Make "n" the interval, in f-numbers, between pitch and
	 '(pitch - pitchRange).
	 n = freqNums((pitch - pitchRange) MOD 12)
	 IF (n > effNbr) THEN n = n \ 2
	 n = effNbr - n
	
	 'Subtract a fraction of the interval to effNbr.
	 effNbr = effNbr - ((n * (&H2000 - vPitchBend(voice))) \ &H2000)
	
	 'If effNbr exceeds its minimum desirable value, bring it into range
	 'by multiplying by two, which raises the pitch by an octave, and
	 'subtract 1 from the octave to keep the pitch in the correct octave.
	 WHILE (effNbr < freqNums(0))
	    effNbr = effNbr * 2
	    octave = octave - 1
	 WEND

      END IF
   END IF

   'Write the lower 8 bits of the f-number.
   CALL SndOutput(&HA0 + voice, effNbr)

   IF (percussion <> 0 AND voice >= BD) THEN
      'Note on/offs for the 5 percussive voices are done using the on/off
      'bits in register &HBD.  See SndSAmVibRhythm().
      t1 = 0
   ELSE
      'ON bit for melodic voices.
      t1 = &H20
   END IF
  
   'Arrange the key-on bit, the octave and the upper two bits of the
   'f-number in the correct order and write to the register.
   t1 = t1 OR (octave * 4) OR (3 AND (effNbr \ 256))
   bxRegister(voice) = t1

   CALL SndOutput(&HB0 + voice, t1)

END SUB

'Used to change the parameter 'param' of the slot 'slot' to 'value'.
'(This routine is not used in this program but may be of use to you.)
'
SUB SetASlotParam (slot AS INTEGER, param AS INTEGER, value AS INTEGER)
   paramSlot(slot, param) = value
   CALL SndSetPrm(slot, param)
END SUB

'Set the 3 global parameters AmDepth, VibDepth and NoteSel.
'
SUB SetGParam (amD AS INTEGER, vibD AS INTEGER, nSel AS INTEGER)
   amDepth = amD
   vibDepth = vibD
   noteSel = nSel

   SndSAmVibRhythm
   SndSNoteSel
END SUB

'Put the chip in melodic mode (mode == 0), or in percussive mode (mode <> 0).
'If the melodic mode is chosen, all voices are set to electric-piano, else
'the first 5 are set to electric-piano, and the percussion voices to their
'default timbres.
'
SUB SetMode (mode AS INTEGER)
   percussion = mode

   IF (mode <> 0) THEN
      modeVoices = 11
   ELSE
      modeVoices = 9
   END IF
   percBits = 0

   IF (mode <> 0) THEN
      'set the frequency for the tom-tom and the snare drum.
      voiceNote(TOM) = TOMPITCH
      vPitchBend(TOM) = NOBEND
      CALL OutFreq(TOM, TOMPITCH)

      voiceNote(SD) = SDPITCH
      vPitchBend(SD) = NOBEND
      CALL OutFreq(SD, SDPITCH)
   END IF
 
   InitSlotParams
   SndSAmVibRhythm
END SUB

'Routine to change the pitch bend range. The value can be from 1 to 12 (in
'half-tones).
'A pitch bend range of 12 means that the pitch bend will be applied to an
'interval of 12 half-tones.  The pitch bend is a value from 0 to &H3fff
'where &H2000 means no change in pitch.  A value above &H2000 means bend up;
'below &H2000, bend down.  The pitch bend is a fraction to be applied to
'the pitch bend range.
SUB SetPitchRange (pR AS INTEGER)
   pitchRange = pR
   IF (pitchRange > 12) THEN
      pitchRange = 12
   ELSEIF (pitchRange < 1) THEN
      pitchRange = 1
   END IF
END SUB

'Set the 14 parameters (13 in 'param', 1 in 'waveSel') of slot 'slot'.
'Update the parameter array and the chip.
'
SUB SetSlotParam (slot AS INTEGER, param() AS INTEGER, waveSel AS INTEGER)
   DIM i AS INTEGER

   FOR i = 0 TO 12
      paramSlot(slot, i) = param(i)
   NEXT
   waveSel = waveSel AND 3
   paramSlot(slot, i) = waveSel
   SndSetAllPrm (slot)
END SUB

'Read the information stored in the DATA statements into the appropriate
'data structures.
'
SUB SetUpData
   DIM n AS INTEGER

   'Read parameters for PIANO instrument, operators 1 and 2
   RESTORE DpianoParamsOp0
   FOR n = 0 TO insParams
      READ pianoParamsOp0(n)
   NEXT
   FOR n = 0 TO insParams
      READ pianoParamsOp1(n)
   NEXT

   'Read parameters for percussive voices
   RESTORE DpercParams
   FOR n = 0 TO insParams
      READ bdOpr0(n)
   NEXT
   FOR n = 0 TO insParams
      READ bdOpr1(n)
   NEXT
   FOR n = 0 TO insParams
      READ sdOpr(n)
   NEXT
   FOR n = 0 TO insParams
      READ tomOpr(n)
   NEXT
   FOR n = 0 TO insParams
      READ cymbOpr(n)
   NEXT
   FOR n = 0 TO insParams
      READ hhOpr(n)
   NEXT
   
   'Read table which gives the voice-slot# correspondance in melodic mode
   RESTORE DslotMVoice
   FOR voice% = 0 TO 8
      FOR slot% = 0 TO 1
	 READ slotMVoice(voice%, slot%)
      NEXT slot%
   NEXT voice%

   'Read table which gives the voice-slot# correspondance in percussive mode
   RESTORE DslotPVoice
   FOR voice% = 0 TO 10
      FOR slot% = 0 TO 1
	 READ slotPVoice(voice%, slot%)
      NEXT slot%
   NEXT voice%

   'Read table of offsets for slots
   RESTORE DoffsetSlot
   FOR n = 0 TO 17
      READ offsetSlot(n)
   NEXT

   'Read table of slot modulator/carrier info
   RESTORE DcarrierSlot
   FOR n = 0 TO 17
      READ carrierSlot(n)
   NEXT

   'Read table which gives the voice number associated with each slot when
   'in melodic mode.
   RESTORE DvoiceMSlot
   FOR n = 0 TO 17
      READ voiceMSlot(n)
   NEXT

   'This table gives the logical voice number  (0-10) associated with
   'each slot (percussive mode only),
   RESTORE DvoicePSlot
   FOR n = 0 TO 17
      READ voicePSlot(n)
   NEXT

   'Read table which gives bit number to use when doing a note on for a
   'percussive voice
   RESTORE DpercMasks
   FOR n = 0 TO 4
      READ percMasks(n)
   NEXT

   'Table of f-numbers to use given a note.
   RESTORE DfreqNums
   FOR n = 0 TO 11
      READ freqNums(n)
   NEXT
     
END SUB

'Do a pitch bend.  (A pitch bend is a small change in pitch.)
'
'The variation in pitch is a function of the previous call to
'SetPitchRange() and the value of pitchBend.
'A pitchBend of 0 means -half-tone * pitchRange, &H2000 means no variation
'(exact pitch) and &H3fff means +half-tone * pitchRange.
'
'Does not affect the percussive voices, except for the bass drum.
'
SUB SetVoicePitch (voice AS INTEGER, pitchBend AS INTEGER)
   IF (voice <= BD OR (percussion <> 0 AND voice < 9)) THEN
      IF (pitchBend > MAXBEND) THEN pitchBend = MAXBEND
      vPitchBend(voice) = pitchBend
      IF (bxRegister(voice) AND &H20) THEN
	 'If note is currently playing, change the pitch.
	 CALL OutFreq(voice, voiceNote(voice))
      END IF
   END IF
END SUB

'Set the instrument (patch) parameters of a voice.
'
'A timbre (melodic or percussive) is defined as follows:
'   13 parameters for operator 0 (ksl, multi, feedBack, attack, sus level,
'       sus sound, decay, release, level, am, vib, ksr, fm)
'   13 parameters for operator 1 (same as for operator 1)
'   Wave form for operator 0
'   Wave form for operator 1
'(This is the same structure as that described in the data section record
'of the .BNK file structure in Appendix B of the Programmer's Manual.)
'
SUB SetVoiceTimbre (voice AS INTEGER, paramArray() AS INTEGER)
   DIM wave0 AS INTEGER
   DIM wave1 AS INTEGER
   DIM slot0 AS INTEGER
   DIM slot1 AS INTEGER
   DIM prm1(0 TO insParams)  AS INTEGER

   IF (voice >= modeVoices) THEN EXIT SUB

   'Determine the slot numbers for the 2 operators of the given voice.
   IF (percussion <> 0) THEN
      slot0 = slotPVoice(voice, 0)
      slot1 = slotPVoice(voice, 1)
   ELSE
      slot0 = slotMVoice(voice, 0)
      slot1 = slotMVoice(voice, 1)
   END IF

   wave0 = paramArray(26)      'wave form for operator 0

   'Write the parameters for operator 0 to the chip.
   CALL SetSlotParam(slot0, paramArray(), wave0)
  
   IF (slot1 <> 255) THEN
      'If here, then this is not a percussive voice and we must set
      'the parameters for the other operator.
      wave1 = paramArray(27)      'wave form for operator 1
      FOR n% = 0 TO insParams
	 'Copy the op1 parameters to a new array.
	 prm1(n%) = paramArray(n% + insParams)
      NEXT

      'Write the parameters for operator 1 to the chip.
      CALL SetSlotParam(slot1, prm1(), wave1)
   END IF
END SUB

'Set the volume of a voice.  The change takes place immediately.
'
SUB SetVoiceVolume (voice AS INTEGER, volume AS INTEGER)
   DIM slot0 AS INTEGER
   DIM slot1 AS INTEGER

   IF (voice >= modeVoices) THEN EXIT SUB

   IF (volume > MAXVOLUME) THEN volume = MAXVOLUME

   lVoiceVolume(voice) = volume

   'Determine the slot numbers for the voice.
   IF (percussion <> 0) THEN
     slot0 = slotPVoice(voice, 0)
     slot1 = slotPVoice(voice, 1)
   ELSE
     slot0 = slotMVoice(voice, 0)
     slot1 = slotMVoice(voice, 1)
   END IF

   'Write new volume to chip.
   SndSKslLevel (slot0)
   IF (slot1 <> 255) THEN SndSKslLevel (slot1)

END SUB

'Set the Wave Select Enable (WSE) bit.  state = 1 enables this feature,
'state = 0 disables.
'
'If you do not want to use the wave-select parameters, call this function
'with a value of 0 AFTER calling SoundColdInit().  By default wave select
'is enabled when the card is initialized.
'
SUB SetWaveSel (state AS INTEGER)
   IF (state <> 0) THEN
      'enable wave select
      modeWaveSel = &H20
   ELSE
      'disable wave select
      modeWaveSel = 0
   END IF

   'Initialize all operators to use the standard sine wave.
   FOR i = 0 TO 17
      CALL SndOutput(&HE0 + offsetSlot(i), 0)
   NEXT
  
   'Set the WSE bit.
   CALL SndOutput(1, modeWaveSel)
END SUB

'Write to the register at &HBD which controls AM Depth, VIB depth, rhythm
'(melo/perc mode) and note on/off for the percussive voices.
'
SUB SndSAmVibRhythm
   DIM t1 AS INTEGER

   t1 = percBits
   IF (amDepth <> 0) THEN t1 = t1 OR &H80
   IF (vibDepth <> 0) THEN t1 = t1 OR &H40
   IF (percussion <> 0) THEN t1 = t1 OR &H20

   CALL SndOutput(&HBD, t1)
END SUB

'Write to the register at &H60 which controls the attack and decay rate.
'
SUB SndSAttDecay (slot AS INTEGER)
   DIM t1 AS INTEGER

   t1 = paramSlot(slot, prmAttack) * 16     '<< 4
   t1 = t1 OR (paramSlot(slot, prmDecay) AND &HF)
   CALL SndOutput(&H60 + offsetSlot(slot), t1)
END SUB

'Write to the register at &H20 which controls AM, VIB, EG-TYP (Sustaining),
'KSR, and MULTI.
'
SUB SndSAVEK (slot AS INTEGER)
   DIM t1 AS INTEGER

   t1 = paramSlot(slot, prmMulti) AND &HF
   IF (paramSlot(slot, prmAm) <> 0) THEN t1 = t1 OR &H80
   IF (paramSlot(slot, prmVib) <> 0) THEN t1 = t1 OR &H40
   IF (paramSlot(slot, prmStaining) <> 0) THEN t1 = t1 OR &H20
   IF (paramSlot(slot, prmKsr) <> 0) THEN t1 = t1 OR &H10

   CALL SndOutput(&H20 + offsetSlot(slot), t1)
END SUB

'Transfer all the parameters for slot 'slot' to the chip.
'
SUB SndSetAllPrm (slot AS INTEGER)
   SndSAmVibRhythm
   SndSNoteSel
   SndSKslLevel (slot)
   SndSFeedFm (slot)
   SndSAttDecay (slot)
   SndSSusRelease (slot)
   SndSAVEK (slot)
   SndWaveSelect (slot)
END SUB

'Update the parameter 'prm' for the slot 'slot'.
'(This routine is not used in this program but may be of use to you.)
'
SUB SndSetPrm (slot AS INTEGER, prm AS INTEGER)
   SELECT CASE (prm)
      CASE prmPercussion, prmAmDepth, prmVibDepth
	 SndSAmVibRhythm
      CASE prmNoteSel
	 SndSNoteSel
      CASE prmKsl, prmLevel
	 SndSKslLevel (slot)
      CASE prmFm, prmFeedBack
	 SndSFeedFm (slot)
      CASE prmAttack, prmDecay
	 SndSAttDecay (slot)
      CASE prmRelease, prmSustain:
	 SndSSusRelease (slot)
      CASE prmMulti, prmVib, prmStaining, prmKsr, prmAm
	 SndSAVEK (slot)
      CASE prmWaveSel
	 SndWaveSelect (slot)
   END SELECT
END SUB

'Write to the register at &HC0 which controls FEED-BACK and FM (connection).
'Applicable only to operator 0.
'
SUB SndSFeedFm (slot AS INTEGER)
   DIM t1 AS INTEGER

   'If this is not an operator 0, return.
   IF (carrierSlot(slot) <> 0) THEN EXIT SUB

   t1 = paramSlot(slot, prmFeedBack) * 2
   IF (paramSlot(slot, prmFm) = 0) THEN t1 = t1 OR 1
   CALL SndOutput(&HC0 + voiceMSlot(slot), t1)
END SUB

'Write to the register at &H40 which controls KSL and output level.
'
SUB SndSKslLevel (slot AS INTEGER)
   DIM level AS INTEGER
   DIM vc AS INTEGER
   DIM flag AS INTEGER

   IF (percussion <> 0) THEN
      vc = voicePSlot(slot)
   ELSE
      vc = voiceMSlot(slot)
   END IF

   'Set level to the output level as specified in the instrument parameters.
   level = 63 - (paramSlot(slot, prmLevel) AND 63)

   'The output volume is the result of modifying the output level for the
   'particular instrument being used by the voice for this slot.
   'For FM instruments, only a modification of operator 0 will result in a
   'change in volume.  (Modifying operator 1 will cause a change of timbre.)
   'For additive synthesis instruments, the output level of both operators
   'are modified to cause a change in volume.
   'For percussive instruments using only 1 operator (i.e. not the BD), we
   'modify the level to change the volume.

   'The following sets flag to 1 if this is the type of operator where we
   'should modify level to cause a change in volume.
   IF (carrierSlot(slot) <> 0 OR paramSlot(slot, prmFm) = 0) THEN
      'carrier (op0) or additive synthesis
      flag = 1
   ELSE
      'percussive voice, not BD
      IF (percussion <> 0 AND vc > BD) THEN flag = 1
   END IF

   IF (flag = 1) THEN
      'level is multiplied by a fraction of the maximum volume.
      level = level * lVoiceVolume(vc) \ 127
   END IF
  
   level = 63 - level
   level = level OR (paramSlot(slot, prmKsl) * 64)      'left shift 6

   CALL SndOutput(&H40 + offsetSlot(slot), level)
END SUB

'Write to the register at &H8.
SUB SndSNoteSel
   IF (noteSel <> 0) THEN
      CALL SndOutput(&H8, 64)
   ELSE
      CALL SndOutput(&H8, 0)
   END IF
END SUB

'Write to the register at &H80 which controls the sustain level and
'release rate.
'
SUB SndSSusRelease (slot AS INTEGER)
   DIM t1 AS INTEGER

   t1 = paramSlot(slot, prmSustain) * 16
   t1 = t1 OR (paramSlot(slot, prmRelease) AND &HF)
   CALL SndOutput(&H80 + offsetSlot(slot), t1)
END SUB

'Write to the register at &HE0 which controls the type of wave-form which
'will be used for this slot.
'
SUB SndWaveSelect (slot AS INTEGER)
   DIM wave AS INTEGER

   IF (modeWaveSel <> 0) THEN
      wave = paramSlot(slot, prmWaveSel)
      wave = wave AND 3
   ELSE
      wave = 0
   END IF
   CALL SndOutput(&HE0 + offsetSlot(slot), wave)
END SUB

'Must be called for start-up initialisation.  Returns 1 if harware found
'else returns 0.
'
FUNCTION SoundColdInit% (port AS INTEGER)
   DIM result AS INTEGER

   genAddr = port
   result = BoardInstalled
   IF (result <> 0) THEN SoundWarmInit
   SoundColdInit = result
END FUNCTION

'-----------------------------------------------
'       Initialize the chip in melodic mode (mode == 0),
'       set all 9 voices to electric-piano timbres,
'       set the 3 global parameters to zero,
'       set the pitch bend range to 1 half-tone,
'       set the pitch bend of each voice to &H2000 (no detune),
'       set the volume of each voice to maximum level,
'       and enable the wave-select parameter.
'-----------------------------------------------
SUB SoundWarmInit
   DIM i AS INTEGER

   SetUpData
  
   FOR i = 1 TO &HF5
      CALL SndOutput(i, 0)    'clear all registers
   NEXT

   CALL SndOutput(4, 6)         'mask T1 & T2

   FOR i = 0 TO 8          'pitch bend for each voice = no detune
      vPitchBend(i) = NOBEND
      voiceNote(i) = 0
   NEXT

   FOR i = 0 TO 10
      lVoiceVolume(i) = MAXVOLUME
   NEXT

   SetMode (0)                 'melodic mode
   CALL SetGParam(0, 0, 0)     'init global parameters
   SetPitchRange (1)           'default pitch range is 1 half-tone
   SetWaveSel (1)              'turn on wave select option
END SUB

'This is a BASIC version of the assembler routine SndOutput.  The purpose of
'the INP instructions is to generate a delay between successive writes to
'the data port.  Because the INP accesses the bus and bus speed is standard
'regardless of processor speed, a delay can easily be generated without
'having to take into account the processor being used.
'However, because of BASIC overhead, the delays generated below will not
'be accurate, so it is best to call the assembler routine (in the library).
SUB xSndOutput (regAddr AS INTEGER, dataByte AS INTEGER)
   DIM i AS INTEGER
  
   OUT &H388, regAddr
   tmp% = INP(&H388)
   tmp% = INP(&H388)
   tmp% = INP(&H388)
   tmp% = INP(&H388)
   tmp% = INP(&H388)
   tmp% = INP(&H388)
  
   OUT &H389, dataByte
   FOR i = 0 TO 36
      tmp% = INP(&H388)
   NEXT
END SUB

